@@ -44,11 +44,11 @@ Test::MockFile - Allows tests to validate code that can interact with files with
 
 =head1 VERSION
 
-Version 0.025
+Version 0.026
 
 =cut
 
-our $VERSION = '0.025';
+our $VERSION = '0.026';
 
 our %files_being_mocked;
 
@@ -167,10 +167,6 @@ sub _upgrade_barewords {
     # Default: no
     unshift @args, 0;
 
-    # Ignore handle objects
-    ref $args[1]
-      and return @args;
-
     # Ignore variables
     # Barewords are provided as strings, which means they're read-only
     # (Of course, readonly scalars here will fool us...)
@@ -275,7 +271,8 @@ sub file {
     ( defined $file && length $file ) or confess("No file provided to instantiate $class");
     _get_file_object($file) and confess("It looks like $file is already being mocked. We don't support double mocking yet.");
 
-    _validate_path($file);
+    my $path = _abs_path_to_file($file);
+    _validate_path($path);
 
     if ( @stats > 1 ) {
         confess(
@@ -285,7 +282,7 @@ sub file {
     }
 
     !defined $contents && @stats
-      and confess("You cannot set stats for non-existent file '$file'");
+      and confess("You cannot set stats for non-existent file '$path'");
 
     my %stats;
     if (@stats) {
@@ -301,14 +298,14 @@ sub file {
     # Check if directory for this file is an object we're mocking
     # If so, mark it now as having content
     # which is this file or - if this file is undef, . and ..
-    ( my $dirname = $file ) =~ s{ / [^/]+ $ }{}xms;
+    ( my $dirname = $path ) =~ s{ / [^/]+ $ }{}xms;
     if ( defined $contents && $files_being_mocked{$dirname} ) {
         $files_being_mocked{$dirname}{'has_content'} = 1;
     }
 
     return $class->new(
         {
-            'path'     => $file,
+            'path'     => $path,
             'contents' => $contents,
             %stats
         }
@@ -366,6 +363,14 @@ sub symlink {
 
     _get_file_object($file) and confess("It looks like $file is already being mocked. We don't support double mocking yet.");
 
+    # Check if directory for this file is an object we're mocking
+    # If so, mark it now as having content
+    # which is this file or - if this file is undef, . and ..
+    ( my $dirname = $file ) =~ s{ / [^/]+ $ }{}xms;
+    if ( $files_being_mocked{$dirname} ) {
+        $files_being_mocked{$dirname}{'has_content'} = 1;
+    }
+
     return $class->new(
         {
             'path'     => $file,
@@ -386,7 +391,7 @@ sub _validate_path {
 
     # Reject the following:
     # ./ ../ /. /.. /./ /../
-    if ( $path =~ m{ ( ^ | / ) \.{1,2} ( / | $ ) }xms ) {
+    if ( $path =~ m{ ( ^ | / ) \.{2} ( / | $ ) }xms ) {
         confess('Relative paths are not supported');
     }
 }
@@ -448,19 +453,21 @@ in the future.)
 =cut
 
 sub dir {
-    my ( $class, $dir_name ) = @_;
+    my ( $class, $dirname ) = @_;
 
-    ( defined $dir_name && length $dir_name ) or confess("No directory name provided to instantiate $class");
-    _get_file_object($dir_name)
-      and confess("It looks like $dir_name is already being mocked. We don't support double mocking yet.");
+    ( defined $dirname && length $dirname ) or confess("No directory name provided to instantiate $class");
+    _get_file_object($dirname)
+      and confess("It looks like $dirname is already being mocked. We don't support double mocking yet.");
 
-    _validate_path($dir_name);
+    my $path = _abs_path_to_file($dirname);
+    _validate_path($path);
 
     # Cleanup trailing forward slashes
-    $dir_name =~ s{[/\\]$}{}xmsg;
+    $path ne '/'
+        and $path =~ s{[/\\]$}{}xmsg;
 
     @_ > 2
-      and confess("You cannot set stats for nonexistent dir '$dir_name'");
+      and confess("You cannot set stats for nonexistent dir '$path'");
 
     my $perms = S_IFPERMS & 0777;
     my %stats = ( 'mode' => ( $perms ^ umask ) | S_IFDIR );
@@ -468,10 +475,10 @@ sub dir {
     # TODO: Add stat information
 
     # FIXME: Quick and dirty: provide a helper method?
-    my $has_content = grep m{^\Q$dir_name/\E}xms, %files_being_mocked;
+    my $has_content = grep m{^\Q$path/\E}xms, %files_being_mocked;
     return $class->new(
         {
-            'path'        => $dir_name,
+            'path'        => $path,
             'has_content' => $has_content,
             %stats
         }
@@ -636,6 +643,8 @@ sub _find_file_or_fh {
 
     # Find the file handle or fall back to just using the abs path of $file_or_fh
     my $absolute_path_to_file = _fh_to_file($file_or_fh) // _abs_path_to_file($file_or_fh) // '';
+    $absolute_path_to_file ne '/'
+        and $absolute_path_to_file =~ s{[/\\]$}{}xmsg;
 
     # Get the pointer to the object.
     my $mock_object = $files_being_mocked{$absolute_path_to_file};
@@ -755,11 +764,18 @@ sub contents {
         # Retrieve the files in this directory and removes prefix
         my $dirname        = $self->path();
         my @existing_files = sort map {
+            # strip directory from the path
             ( my $basename = $_->path() ) =~ s{^\Q$dirname/\E}{}xms;
-            defined $_->{'contents'} ? ($basename) : ();
+
+            # Is this content within another directory? strip that out
+            $basename =~ s{^( [^/]+ ) / .*}{$1}xms;
+
+            defined $_->{'contents'} || $_->is_link() || $_->is_dir() ? ($basename) : ();
         } _files_in_dir($dirname);
 
-        return [ '.', '..', @existing_files ];
+        my %uniq;
+        $uniq{$_}++ for @existing_files;
+        return [ '.', '..', sort keys %uniq ];
     }
 
     # handle files
@@ -1240,7 +1256,7 @@ BEGIN {
     *CORE::GLOBAL::open = sub(*;$@) {
         my $likely_bareword;
         my $arg0;
-        if ( defined $_[0] ) {
+        if ( defined $_[0] && !ref $_[0] ) {
 
             # We need to remember the first arg to override the typeglob for barewords
             $arg0 = $_[0];
@@ -1431,7 +1447,7 @@ BEGIN {
     *CORE::GLOBAL::opendir = sub(*$) {
 
         # Upgrade but ignore bareword indicator
-        ( undef, @_ ) = _upgrade_barewords(@_) if defined $_[0];
+        ( undef, @_ ) = _upgrade_barewords(@_) if defined $_[0] && !ref $_[9];
 
         my $mock_dir = _get_file_object( $_[1] );
 
@@ -1480,7 +1496,7 @@ BEGIN {
     *CORE::GLOBAL::readdir = sub(*) {
 
         # Upgrade but ignore bareword indicator
-        ( undef, @_ ) = _upgrade_barewords(@_) if defined $_[0];
+        ( undef, @_ ) = _upgrade_barewords(@_) if defined $_[0] && !ref $_[9];
 
         my $mocked_dir = _get_file_object( $_[0] );
 
@@ -1520,7 +1536,7 @@ BEGIN {
     *CORE::GLOBAL::telldir = sub(*) {
 
         # Upgrade but ignore bareword indicator
-        ( undef, @_ ) = _upgrade_barewords(@_) if defined $_[0];
+        ( undef, @_ ) = _upgrade_barewords(@_) if defined $_[0] && !ref $_[9];
 
         my ($fh) = @_;
         my $mocked_dir = _get_file_object($fh);
@@ -1546,7 +1562,7 @@ BEGIN {
     *CORE::GLOBAL::rewinddir = sub(*) {
 
         # Upgrade but ignore bareword indicator
-        ( undef, @_ ) = _upgrade_barewords(@_) if defined $_[0];
+        ( undef, @_ ) = _upgrade_barewords(@_) if defined $_[0] && !ref $_[9];
 
         my ($fh) = @_;
         my $mocked_dir = _get_file_object($fh);
@@ -1573,7 +1589,7 @@ BEGIN {
     *CORE::GLOBAL::seekdir = sub(*$) {
 
         # Upgrade but ignore bareword indicator
-        ( undef, @_ ) = _upgrade_barewords(@_) if defined $_[0];
+        ( undef, @_ ) = _upgrade_barewords(@_) if defined $_[0] && !ref $_[9];
 
         my ( $fh, $goto ) = @_;
         my $mocked_dir = _get_file_object($fh);
@@ -1599,7 +1615,7 @@ BEGIN {
     *CORE::GLOBAL::closedir = sub(*) {
 
         # Upgrade but ignore bareword indicator
-        ( undef, @_ ) = _upgrade_barewords(@_) if defined $_[0];
+        ( undef, @_ ) = _upgrade_barewords(@_) if defined $_[0] && !ref $_[9];
 
         my ($fh) = @_;
         my $mocked_dir = _get_file_object($fh);
